import 'dart:math';

import 'package:minehunter/entities/cell.dart';
import 'package:minehunter/provider/data.dart';

class BoardUtils {
  static printGrid(List cells) {
    const String isMined = 'X';
    const String isSafe = '.';

    const String mineFound = '#';
    const String wrongMarkedCell = '0';
    const String exploredSafeCell = '.';
    const String unkownState = ' ';

    print('');
    String line = '--';
    for (int i = 0; i < cells[0].length; i++) {
      line += '-';
    }
    print('$line  $line');
    for (int rowIndex = 0; rowIndex < cells.length; rowIndex++) {
      String currentLine = '';
      String solvedLine = '';
      for (int colIndex = 0; colIndex < cells[rowIndex].length; colIndex++) {
        solvedLine += cells[rowIndex][colIndex].isMined ? isMined : isSafe;

        String cellString = unkownState;
        if (cells[rowIndex][colIndex].isExplored) {
          cellString = exploredSafeCell;
        }
        if (cells[rowIndex][colIndex].isMarked) {
          if (cells[rowIndex][colIndex].isMined) {
            cellString = mineFound;
          } else {
            cellString = wrongMarkedCell;
          }
        }
        currentLine += cellString;
      }
      print('|$currentLine|  |$solvedLine|');
    }
    print('$line  $line');
    print('');
  }

  static List<List<Cell>> createEmptyBoard(int sizeHorizontal, int sizeVertical) {
    final List<List<Cell>> cells = [];
    for (int rowIndex = 0; rowIndex < sizeVertical; rowIndex++) {
      final List<Cell> row = [];
      for (int colIndex = 0; colIndex < sizeHorizontal; colIndex++) {
        row.add(Cell(false));
      }
      cells.add(row);
    }

    return cells;
  }

  static int getMinesCount(int sizeHorizontal, int sizeVertical, String level) {
    int minesCountRatio = 0;
    switch (level) {
      case 'easy':
        {
          minesCountRatio = 5;
        }
        break;
      case 'medium':
        {
          minesCountRatio = 10;
        }
        break;
      case 'hard':
        {
          minesCountRatio = 15;
        }
        break;
      case 'nightmare':
        {
          minesCountRatio = 20;
        }
        break;
    }

    final int minesCount = ((sizeHorizontal * sizeVertical) * minesCountRatio / 100).round();

    print('Mines count: $minesCount');

    return minesCount;
  }

  static void createInitialEmptyBoard(Data myProvider) {
    myProvider.updateIsBoardMined(false);
    myProvider
        .updateCells(createEmptyBoard(myProvider.sizeHorizontal, myProvider.sizeVertical));
  }

  static List<List<Cell>> createBoard(Data myProvider, int forbiddenRow, int forbiddenCol) {
    final List<List<Cell>> cells = myProvider.cells;
    final int sizeHorizontal = myProvider.sizeHorizontal;
    final int sizeVertical = myProvider.sizeVertical;

    // Shuffle cells to put random mines, except on currently selected one
    final List<List<int>> allowedCells = [];
    for (int row = 0; row < sizeVertical; row++) {
      for (int col = 0; col < sizeHorizontal; col++) {
        if (!((forbiddenRow == row) && (forbiddenCol == col))) {
          allowedCells.add([row, col]);
        }
      }
    }
    allowedCells.shuffle();

    // Put random mines on board
    for (int mineIndex = 0; mineIndex < myProvider.minesCount; mineIndex++) {
      cells[allowedCells[mineIndex][0]][allowedCells[mineIndex][1]].isMined = true;
    }

    // Compute all mines counts on cells
    for (int row = 0; row < sizeVertical; row++) {
      for (int col = 0; col < sizeHorizontal; col++) {
        cells[row][col].minesCountAround = getMinesCountAround(cells, row, col);
      }
    }

    printGrid(cells);

    return cells;
  }

  static List<List<Cell>> createBoardFromSavedState(Data myProvider, String savedBoard) {
    final List<List<Cell>> board = [];
    final int boardSize = pow((savedBoard.length / 6), 1 / 2).round();
    final String boardSizeAsString = '${boardSize}x$boardSize';
    myProvider.updateParameterSize(boardSizeAsString);

    int index = 0;
    for (int rowIndex = 0; rowIndex < boardSize; rowIndex++) {
      final List<Cell> row = [];
      for (int colIndex = 0; colIndex < boardSize; colIndex++) {
        final bool isMined = (savedBoard[index++] == 'X');
        final bool isExplored = (savedBoard[index++] == 'E');
        final bool isMarked = (savedBoard[index++] == 'P');
        final bool isExploded = (savedBoard[index++] == '*');
        final int minesCountAround = int.parse(savedBoard[index++]);
        index++; // ";"

        Cell cell = Cell(isMined);
        cell.isExplored = isExplored;
        cell.isMarked = isMarked;
        cell.isExploded = isExploded;
        cell.minesCountAround = minesCountAround;

        row.add(cell);
      }
      board.add(row);
    }

    printGrid(board);

    return board;
  }

  static void reportCell(Data myProvider, int row, int col) {
    if (!myProvider.cells[row][col].isExplored) {
      myProvider.toggleCellMark(row, col);
    }
  }

  static void walkOnCell(Data myProvider, int row, int col) {
    myProvider.setCellAsExplored(row, col);

    if (myProvider.cells[row][col].minesCountAround == 0) {
      final List<List<int>> safeCells = getAllSafeCellsAround(myProvider.cells, row, col);
      for (int safeCellIndex = 0; safeCellIndex < safeCells.length; safeCellIndex++) {
        final int safeCellRow = safeCells[safeCellIndex][0];
        final int safeCellCol = safeCells[safeCellIndex][1];
        if (!myProvider.cells[safeCellRow][safeCellCol].isExplored) {
          walkOnCell(myProvider, safeCellRow, safeCellCol);
        }
      }
    }
  }

  static List<List<int>> getAllSafeCellsAround(List cells, int row, int col) {
    final int sizeHorizontal = cells.length;
    final int sizeVertical = cells[0].length;

    final List<List<int>> safeCellsCoordinates = [];

    if (cells[row][col].minesCountAround == 0) {
      for (int deltaRow = -1; deltaRow <= 1; deltaRow++) {
        for (int deltaCol = -1; deltaCol <= 1; deltaCol++) {
          final int candidateRow = row + deltaRow;
          final int candidateCol = col + deltaCol;
          if ((candidateRow >= 0 && candidateRow < sizeVertical) &&
              (candidateCol >= 0 && candidateCol < sizeHorizontal) &&
              !cells[candidateRow][candidateCol].isExplored) {
            safeCellsCoordinates.add([candidateRow, candidateCol]);
          }
        }
      }
    }

    return safeCellsCoordinates;
  }

  static int getMinesCountAround(List cells, int row, int col) {
    final int sizeHorizontal = cells.length;
    final int sizeVertical = cells[0].length;

    int minesCountAround = 0;
    for (int deltaRow = -1; deltaRow <= 1; deltaRow++) {
      for (int deltaCol = -1; deltaCol <= 1; deltaCol++) {
        if ((row + deltaRow >= 0 && row + deltaRow < sizeVertical) &&
            (col + deltaCol >= 0 && col + deltaCol < sizeHorizontal) &&
            (cells[row + deltaRow][col + deltaCol].isMined)) {
          minesCountAround++;
        }
      }
    }

    return minesCountAround;
  }

  static bool checkGameIsFinished(Data myProvider) {
    final List<List<Cell>> cells = myProvider.cells;
    final int sizeHorizontal = cells.length;
    final int sizeVertical = cells[0].length;

    printGrid(cells);

    myProvider.updateGameWin(false);
    myProvider.updateGameFail(false);

    for (int row = 0; row < sizeVertical; row++) {
      for (int col = 0; col < sizeHorizontal; col++) {
        // Walked on a mine
        if (cells[row][col].isExploded == true) {
          myProvider.updateGameFail(true);
          return true;
        }
      }
    }

    for (int row = 0; row < sizeVertical; row++) {
      for (int col = 0; col < sizeHorizontal; col++) {
        if (
            // Mine not already found
            (cells[row][col].isMined == true && cells[row][col].isMarked == false) ||
                // Safe cell marked as mined
                (cells[row][col].isMined == false && cells[row][col].isMarked == true)) {
          return false;
        }
      }
    }

    print('-> ok all mines found!');
    myProvider.updateGameWin(true);

    return true;
  }

  static int countFlaggedCells(List cells) {
    int count = 0;

    final int sizeHorizontal = cells.length;
    final int sizeVertical = cells[0].length;
    for (int row = 0; row < sizeVertical; row++) {
      for (int col = 0; col < sizeHorizontal; col++) {
        if (cells[row][col].isMarked == true) {
          count++;
        }
      }
    }

    return count;
  }
}
